#ifndef MUBLAS_H
#define MUBLAS_H

#include <cuComplex.h>

// ***** FORMAT *****

struct __align__(8) cudfreal
{
	float x;
	float y;
};

struct __align__(16) cuddreal
{
	double x;
	double y;
};

struct __align__(16) cudfcomplex
{
	cudfreal x;
	cudfreal y;
};

struct cuddcomplex
{
	cuddreal x;
	cuddreal y;
};

// cudfreal ----------------------------------
__host__ __device__ __forceinline__ cudfreal
make_cudfreal 
	(const float h, const float l)
{
	cudfreal v;
	v.x = h;
	v.y = l;
	return v;
}
// cudfreal ----------------------------------

// cudfcomplex ----------------------------------
__host__ __device__ __forceinline__ cudfcomplex
make_cudfcomplex 
	(const cudfreal re, const cudfreal im)
{
	cudfcomplex v;
	v.x = re;
	v.y = im;
	return v;
}

__host__ __device__ __forceinline__ cudfcomplex
make_cudfcomplex
	(const float reh, const float rel, const float imh, const float iml)
{
	cudfcomplex v;
	v.x.x = reh;
	v.x.y = rel;
	v.y.x = imh;
	v.y.y = iml;
	return v;
}
// cudfcomplex ----------------------------------

// cuddreal ----------------------------------
__host__ __device__ __forceinline__ cuddreal
make_cuddreal 
	(const double h, const double l)
{
	cuddreal v;
	v.x = h;
	v.y = l;
	return v;
}

__host__ __device__ __forceinline__ cuddreal
make_cuddreal 
	(const double h)
{
	cuddreal v;
	v.x = h;
	v.y = 0.;
	return v;
}
// cuddreal ----------------------------------

// cuddcomplex ----------------------------------
__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex 
	(const cuddreal re, const cuddreal im)
{
	cuddcomplex v;
	v.x = re;
	v.y = im;
	return v;
}

__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex 
	(const double re, const double im)
{
	cuddcomplex v;
	v.x.x = re;
	v.x.y = 0.;
	v.y.x = im;
	v.y.y = 0.;
	return v;
}

__host__ __device__ __forceinline__ cuddcomplex
make_cuddcomplex
	(const double reh, const double rel, const double imh, const double iml)
{
	cuddcomplex v;
	v.x.x = reh;
	v.x.y = rel;
	v.y.x = imh;
	v.y.y = iml;
	return v;
}
// cuddcomplex ----------------------------------

struct mublasHandle_t {
	cudaDeviceProp devprop;  
	cudaFuncAttributes funcatb;
	int32_t driverVersion; 
	int32_t runtimeVersion;
	int32_t cc;

	// kernel parameters (should be stored in another handler...?)
	int32_t kernelnumber;	// Kernel Number

	int32_t Tx_B;			// # of threads for X-dimention per Block
	int32_t Ty_B;			// # of threads for Y-dimention per Block
	int32_t Tx_B_static;
	int32_t Ty_B_static;

	int32_t Tx_B_min;
	int32_t B_M_min;
	int32_t B_M_optmin;

	int64_t m;
	int64_t n;
	int32_t N_one;			// Bytes of One Element [Bytes]
	int32_t N_T;
	int32_t USEvector;
	int32_t S_B;
	int32_t B_D;
	int32_t B_M_max;
	int32_t W_M_pmax;		// Physical Max of Warps per Multiprocessor
	int32_t B_M_pmax;		// Physical Max of Blocks per Multiprocessor
	int32_t R_unit;			// Register Allocation Unit
	int32_t S_unit;			// Shared Memory Allocation Unit
	int32_t W_gran;			// Warp Allocation Granularity
	int32_t Cores;			// # of CUDA Cores per Multiprocessor
	int32_t DPCores;		// # of DP Cores per Multiprocessor
	float BO;				// Block Occupancy
	float WO;				// Warp Occupancy
	float BO2;				// debug
};

void
mublasCreate (mublasHandle_t*);

// GEMV
extern int32_t mublasSgemv (mublasHandle_t*, char, const int32_t, const int32_t, const float*, const float*, const int32_t, const float*, const int32_t, const float*, float*, const int32_t);
extern int32_t mublasDgemv (mublasHandle_t*, char, const int32_t, const int32_t, const double*, const double*, const int32_t, const double*, const int32_t, const double*, double*, const int32_t);
extern int32_t mublasCgemv (mublasHandle_t*, char, const int32_t, const int32_t, const cuFloatComplex*, const cuFloatComplex*, const int32_t, const cuFloatComplex*, const int32_t, const cuFloatComplex*, cuFloatComplex*, const int32_t);
extern int32_t mublasZgemv (mublasHandle_t*, char, const int32_t, const int32_t, const cuDoubleComplex*, const cuDoubleComplex*, const int32_t, const cuDoubleComplex*, const int32_t, const cuDoubleComplex*, cuDoubleComplex*, const int32_t);
extern int32_t mublasWgemv (mublasHandle_t*, char, const int32_t, const int32_t, const cuddreal*, const cuddreal*, const int32_t, const cuddreal*, const int32_t, const cuddreal*, cuddreal*, const int32_t);
extern int32_t mublasUgemv (mublasHandle_t*, char, const int32_t, const int32_t, const cuddcomplex*, const cuddcomplex*, const int32_t, const cuddcomplex*, const int32_t, const cuddcomplex*, cuddcomplex*, const int32_t);
extern int32_t mublasBgemv (mublasHandle_t*, char, const int32_t, const int32_t, const cudfreal*, const cudfreal*, const int32_t, const cudfreal*, const int32_t, const cudfreal*, cudfreal*, const int32_t);
extern int32_t mublasVgemv (mublasHandle_t*, char, const int32_t, const int32_t, const cudfcomplex*, const cudfcomplex*, const int32_t, const cudfcomplex*, const int32_t, const cudfcomplex*, cudfcomplex*, const int32_t);

#endif
