// ***** SETTING ***********************************

// octuple-precision check using MPACK/MBLAS (set "(use_mblas TRUE)" in CMakeLists.txt)
//#define MBLAS_CHECK

// enable temperature measurement using NVIDIA Management Library (NVML) (install and set PATH in CMakeLists.txt)
//#define USE_NVML

// GPU device number (default: 0)
#define DEVNO 0 

// performance measurement 
// (recommended to enable for accurate measurement) a routine is repeatedly executed at least MINSEC second, at least MINLOOP, and at maximum MAXLOOP times for each problem size. When PERF_1SEC is not defined, a routine is executed at least MINLOOP times and then the program displays the best performance among MINLOOP executions
// ############################
#define PERF_1SEC
#define MINSEC 0.1
#define MINLOOP 10
#define MAXLOOP 10000
// ############################

// Experimental
//#define NODISPINFO // does not display any information
#define MEMMARGIN 150000000 // memory margin for calculating maximum problem size
#define TEMPLIMIT 90  // temperature limit 
#define PITCH // pitched alloc (default: yes)
#define DDMADCOUNT 16 // num of double-precision floating instructions for DD-MAD (for calculate theoretical peak performance)
#define DFMADCOUNT 16 // num of double-precision floating instructions for DF-MAD (for calculate theoretical peak performance)
// ***** SETTING ***********************************

// header ------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <sys/time.h>
#include <unistd.h>
#ifndef MBLAS_CHECK
#include <f2c.h>
#endif
#include <algorithm>
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <math.h>
#include <cblas.h>
#include <cublas_v2.h>
#include <getopt.h>
#include <ctype.h>

#ifdef USE_NVML
#include <nvml.h>
#endif

#ifdef MBLAS_CHECK
#include <qd/qd_real.h>
#include <mblas_qd.h>
#define REFERENCE_TYPE MBLAS_TYPE
#define REFERENCE_BASETYPE MBLAS_BASETYPE
#define REFERENCE_ZERO MBLAS_ZERO
#define REFERENCE_ONE MBLAS_ONE
#define REFERENCE_NEGONE MBLAS_NEGONE
#define REFERENCE_AXPY MBLAS_AXPY
#define REFERENCE_NRM MBLAS_NRM
#else
#include <clapack.h>
#define REFERENCE_TYPE CBLAS_TYPE
#define REFERENCE_BASETYPE CBLAS_BASETYPE
#define REFERENCE_ZERO CBLAS_ZERO
#define REFERENCE_ONE CBLAS_ONE
#define REFERENCE_NEGONE CBLAS_NEGONE
#define REFERENCE_AXPY CBLAS_AXPY
#define REFERENCE_NRM CBLAS_NRM
#endif

#include <mublas.h>
#ifdef MUBLAS
#elif CUBLAS
#elif KBLAS
//#include <kblas.h> // kblas.h has no extern
extern "C" int32_t kblas_sgemv (char, int32_t, int32_t, float, float*, int32_t, float*, int32_t, float, float*, int32_t);
extern "C" int32_t kblas_dgemv (char, int32_t, int32_t, double, double*, int32_t, double*, int32_t, double, double*, int32_t);
extern "C" int32_t kblas_cgemv (char, int32_t, int32_t, cuFloatComplex, cuFloatComplex*, int32_t, cuFloatComplex*, int32_t, cuFloatComplex, cuFloatComplex*, int32_t);
extern "C" int32_t kblas_zgemv (char, int32_t, int32_t, cuDoubleComplex, cuDoubleComplex*, int32_t, cuDoubleComplex*, int32_t, cuDoubleComplex, cuDoubleComplex*, int32_t);
extern "C" int32_t kblas_ssymv (char, int32_t, float, float*, int32_t, float*, int32_t, float, float*, int32_t);
extern "C" int32_t kblas_dsymv (char, int32_t, double, double*, int32_t, double*, int32_t, double, double*, int32_t);
extern "C" int32_t kblas_chemv (char, int32_t, cuFloatComplex, cuFloatComplex*, int32_t, cuFloatComplex*, int32_t, cuFloatComplex, cuFloatComplex*, int32_t);
extern "C" int32_t kblas_zhemv (char, int32_t, cuDoubleComplex, cuDoubleComplex*, int32_t, cuDoubleComplex*, int32_t, cuDoubleComplex, cuDoubleComplex*, int32_t);
#elif MAGMA
#include <magma.h>
#elif CULA
#include <cula_blas_device.h>
#elif ASPEN
#include <aspen.h>
#endif

#include "prec_s.h"
#include "prec_d.h"
#include "prec_c.h"
#include "prec_z.h"
#include "prec_w.h"
#include "prec_u.h"
#include "prec_b.h"
#include "prec_v.h"

#ifdef GEMM
#define ROUTINE_FLOPS(m,n,k)			(FPMADCOUNT*m*n*k+(2*FPMULCOUNT+FPADDCOUNT)*m*n)
#define ROUTINE_BYTES(m,n,k,s)			((m*k+k*n+2*m*n)*s) 
#define ROUTINE_MAXSIZE(gmem,margin,s)	(sqrt((gmem-margin)/s/3)) 
#endif

#ifdef GEMV
#define ROUTINE_FLOPS(m,n,k)						(FPMADCOUNT*m*n+2*FPMULCOUNT*m+FPADDCOUNT*m+(k*0))
#define ROUTINE_BYTES(m,n,k,s)						((2*m+n+m*n)*s+(k*0)) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#endif

#ifdef SYMV
#define ROUTINE_FLOPS(m,n,k)						(FPMADCOUNT*n*n+2*FPMULCOUNT*n+FPADDCOUNT*n+(m*k*0))
#define ROUTINE_BYTES(m,n,k,s)						symv_mem(n)*s+(m*k*0) //((3*n+n*n)*s+(m*k*0)) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#endif

#ifdef HEMV
#define ROUTINE_FLOPS(m,n,k)						(FPMADCOUNT*n*n+2*FPMULCOUNT*n+FPADDCOUNT*n+(m*k*0))
#define ROUTINE_BYTES(m,n,k,s)						symv_mem(n)*s+(m*k*0) //((3*n+n*n)*s+(m*k*0)) 
#define ROUTINE_MAXSIZE(gmem,margin,s)				(sqrt((gmem-margin)/s))
#define ROUTINE_MAXSIZE_1(gmem,margin,s,dim1)		((gmem-margin)/s/dim1)
#define ROUTINE_MAXSIZE_2(gmem,margin,s,dim1,dim2)	((gmem+margin+s+dim1+dim2)*0) 
#endif


int64_t symv_mem (int64_t n) {
	int64_t i, v = 0;
	for (i = 0; i < n; i++) 
		v += (n - i);
	v += 3 * n;
	return v;
}

#ifdef AXPY
#define ROUTINE_FLOPS(m,n,k)			(FPMADCOUNT*n+(k+m)*0)
#define ROUTINE_BYTES(m,n,k,s)			(3*n*s+(k+m)*0)
#define ROUTINE_MAXSIZE(gmem,margin,s)	((gmem-margin)/s/m/2+(n+k)*0)
#endif

#ifdef SCAL
#define ROUTINE_FLOPS(m,n,k)			(FPMULCOUNT*n+(k+m)*0)
#define ROUTINE_BYTES(m,n,k,s)			(2*n*s+(m+k)*0)
#define ROUTINE_MAXSIZE(gmem,margin,s)	((gmem-margin)/s/m+(n+k)*0)
#endif

#ifdef COPY
#define ROUTINE_FLOPS(m,n,k)			((m+n+k)*0)
#define ROUTINE_BYTES(m,n,k,s)			(2*n*s+(m+k)*0)
#define ROUTINE_MAXSIZE(gmem,margin,s)	((gmem-margin)/s/m/2+(n+k)*0)
#endif

// header ------------------------------------------

// ----- CUBLAS const value function ---------------------------------------------------------------
inline cublasOperation_t cublas_trans_const_m (char tran) {
	if (tran == 'N' || tran == 'n') return CUBLAS_OP_N;
	if (tran == 'T' || tran == 't') return CUBLAS_OP_T;
	if (tran == 'C' || tran == 'c') return CUBLAS_OP_C;
	return CUBLAS_OP_N; //default
}
inline cublasFillMode_t cublas_uplo_const_m (char uplo) {
	if (uplo == 'L' || uplo == 'l') return CUBLAS_FILL_MODE_LOWER;
	if (uplo == 'U' || uplo == 'u') return CUBLAS_FILL_MODE_UPPER;
	return CUBLAS_FILL_MODE_LOWER; //default
}
inline cublasDiagType_t cublas_diag_const_m (char diag) {
	if (diag == 'N' || diag == 'n') return CUBLAS_DIAG_NON_UNIT;
	if (diag == 'U' || diag == 'u') return CUBLAS_DIAG_UNIT;
	return CUBLAS_DIAG_NON_UNIT; //default
}
inline cublasSideMode_t cublas_side_const_m (char side) {
	if (side == 'L' || side == 'l') return CUBLAS_SIDE_LEFT;
	if (side == 'R' || side == 'r') return CUBLAS_SIDE_RIGHT;
	return CUBLAS_SIDE_LEFT; //default
}

// ----- CBLAS const value function ---------------------------------------------------------------
inline CBLAS_TRANSPOSE cblas_trans_const_m (char tran) {
	if (tran == 'N' || tran == 'n') return CblasNoTrans;
	if (tran == 'T' || tran == 't') return CblasTrans;
	if (tran == 'C' || tran == 'c') return CblasConjTrans;
	return CblasNoTrans; //default
}
inline CBLAS_UPLO cblas_uplo_const_m (char uplo) {
	if (uplo == 'L' || uplo == 'l') return CblasLower;
	if (uplo == 'U' || uplo == 'u') return CblasUpper;
	return CblasLower; //default
}
inline CBLAS_DIAG cblas_diag_const_m (char diag) {
	if (diag == 'N' || diag == 'n') return CblasNonUnit;
	if (diag == 'U' || diag == 'u') return CblasUnit;
	return CblasNonUnit; //default
}
inline CBLAS_SIDE cblas_side_const_m (char side) {
	if (side == 'L' || side == 'l') return CblasLeft;
	if (side == 'R' || side == 'r') return CblasRight;
	return CblasLeft; //default
}

// ----- MBLAS compatible ---------------------------------------------------------------
#ifdef MBLAS_CHECK
inline cuddreal make_cuddreal (const qd_real v) {
	return make_cuddreal (v.x[0], v.x[1]);
}
inline cuddcomplex make_cuddcomplex (const qd_real r, const qd_real i) {
	return make_cuddcomplex (make_cuddreal (r), make_cuddreal (i));
}
#endif
// ----- MBLAS compatible ---------------------------------------------------------------

// ----- testing handler ------------------------------------------------------------------------
struct testingHandle_t {
	// device parameter
	cudaDeviceProp devprop;  
	int32_t cc;
	double device_bytes_per_sec;
	double device_flops_per_sec_sp;
	double device_flops_per_sec_dp;

	int32_t Cores;
	int32_t DPCores;

	// testing parameter
	char mode; // c=check, p=perf

	double routine_flops_per_sec_theoretical_peak;
	double routine_flops;
	double routine_bytes;

	int32_t word_size;
	int32_t Tx_B_static;
	int32_t Ty_B_static;
	int32_t driverVersion; 
	int32_t runtimeVersion;
	int64_t test_points;
	int32_t kernelnumber;
	int32_t cublasatomic;

	// BLAS parameters
	// dim
	int64_t kl;
	int64_t ku;
	int64_t dim_start;
	int64_t dim_stop;
	int64_t dim_step;
	int64_t dim_m;
	int64_t dim_n;
	int64_t dim_k;
	int64_t incx;
	int64_t incy;

	int32_t const_m;
	int32_t const_n;
	int32_t const_k;

	// options
	char tranA; // N, T, C
	char tranB; // N, T, C
	char uplo; // U, L
	char diag; // N, U
	char side; // L, R
};

void
testingCreate (testingHandle_t* th) {
    int driverVersion = 0;
	int runtimeVersion = 0;
	cudaDeviceProp devprop;  
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);
	cudaGetDeviceProperties (&devprop, 0);

	th->devprop = devprop;
	th->driverVersion = driverVersion;
	th->runtimeVersion = runtimeVersion;
	th->cc = devprop.major * 100 + devprop.minor * 10;

	// default ---------------------------
	th->mode = 'c';
	// options
	th->tranA = 'N';
	th->tranB = 'N';
	th->uplo = 'U';
	th->diag = 'N';
	th->side = 'L';

	th->Tx_B_static = 0;
	th->Ty_B_static = 0;
	th->kernelnumber = 0;
	th->cublasatomic = 0;

	th->dim_start	= 128;
	th->dim_stop	= 1024;
	th->dim_step	= 128;
	th->dim_m = 0;
	th->dim_n = 0;
	th->dim_k = 0;
	th->const_m = 0;
	th->const_n = 0;
	th->const_k = 0;
	th->incx = 1;
	th->incy = 1;
	th->test_points = 0;

	switch (th->cc) {
		case 100:
			th->Cores = 8;
			th->DPCores = 1;
			break;
		case 110:
			th->Cores = 8;
			th->DPCores = 1;
			break;
		case 120:
			th->Cores = 8;
			th->DPCores = 1;
			break;
		case 130:
			th->Cores = 8;
			th->DPCores = 1;
			break;
		case 200:
			th->Cores = 32;
			th->DPCores = 16;
			break;
		case 210:
			th->Cores = 48;
			th->DPCores = 0; // ?
			break;
		case 300:
			th->Cores = 192;
			th->DPCores = 64;
			break;
		case 350:
			th->Cores = 192;
			th->DPCores = 64;
			break;
		case 500:
			th->Cores = 128;
			th->DPCores = 0;
			break;
		case 520:
			th->Cores = 128;
			th->DPCores = 4;
			break;
		default: // CC350 
			th->Cores = 192;
			th->DPCores = 64;
			break;
	}

	th->device_bytes_per_sec = (double)th->devprop.memoryClockRate * 2. * 1000. * (double)th->devprop.memoryBusWidth / 8.;
	th->device_flops_per_sec_sp = (double)th->devprop.clockRate * 1000. * 2. * (double)th->devprop.multiProcessorCount * th->Cores;
	th->device_flops_per_sec_dp = (double)th->devprop.clockRate * 1000. * 2. * (double)th->devprop.multiProcessorCount * th->DPCores;
#if defined (PREC_S) || defined (PREC_C) // single precision
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_sp;
#elif defined (PREC_D) || defined (PREC_Z) // double precision
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_dp;
#elif defined (PREC_B) || defined (PREC_DF) || defined (PREC_V) // DF-operation
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_sp / (double)DFMADCOUNT;
#elif defined (PREC_W) || defined (PREC_U) // DD-operation
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_dp / (double)DDMADCOUNT;
#else // unknown
	th->routine_flops_per_sec_theoretical_peak = 0.;
#endif
}

void get_routine_theoretial_performance (testingHandle_t* th) {
	int64_t m, n, k, s;
	m = th->dim_m;
	n = th->dim_n;
	k = th->dim_k;
	s = th->word_size;
	th->routine_flops = (double) ROUTINE_FLOPS (m, n, k);
	th->routine_bytes = (double) ROUTINE_BYTES (m, n, k, s);
}

void dim_initialize (testingHandle_t* th) {
	if (th->const_m == 0) th->dim_m = th->dim_stop;
	if (th->const_n == 0) th->dim_n = th->dim_stop;
	if (th->const_k == 0) th->dim_k = th->dim_stop;
}

void dim_start (testingHandle_t* th) {
	if (th->const_m == 0) th->dim_m = th->dim_start;
	if (th->const_n == 0) th->dim_n = th->dim_start;
	if (th->const_k == 0) th->dim_k = th->dim_start;
}

void dim_step (testingHandle_t* th) {
	if (th->const_m == 0) {
		if (th->dim_step == -222) th->dim_m *= 2; // log mode
		else th->dim_m += th->dim_step;
	}
	if (th->const_n == 0) {
		if (th->dim_step == -222) th->dim_n *= 2; // log mode
		else th->dim_n += th->dim_step;
	}
	if (th->const_k == 0) {
		if (th->dim_step == -222) th->dim_k *= 2; // log mode
		else th->dim_k += th->dim_step;
	}
}

void set_testing_handler (int32_t argc, char **argv, testingHandle_t* th)
{
	int32_t opt, opt_index;
	optind = 1;
	struct option long_options[] = {
		{"m", required_argument, NULL, 'm'},
		{"n", required_argument, NULL, 'n'},
		{"k", required_argument, NULL, 'k'},
		{"range", required_argument, NULL, 'r'},
		{"trana", required_argument, NULL, 'a'},
		{"tranb", required_argument, NULL, 'b'},
		{"uplo", required_argument, NULL, 'u'},
		{"diag", required_argument, NULL, 'd'},
		{"side", required_argument, NULL, 's'},
		{"mode", required_argument, NULL, 'o'},
		{"kernel", required_argument, NULL, 'v'},
		{"atomic", no_argument, NULL, 't'},
		{"tx", required_argument, NULL, 'x'},
		{"ty", required_argument, NULL, 'y'},
		{"points", required_argument, NULL, 'p'},
		{"help", no_argument, NULL, 'h'},
		{"incx", required_argument, NULL, 'i'},
		{"incy", required_argument, NULL, 'j'},
		{0, 0, 0, 0}
	};
	while ((opt = getopt_long (argc, argv, "mnkrabudsovtxyrphij", long_options, &opt_index)) != -1){
		switch (opt) {
			// dim
			case 'm':{
				th->dim_m = (int64_t)atoi(optarg);
				th->const_m = 1;
				break;
			}
			case 'n':{
				th->dim_n = (int64_t)atoi(optarg);
				th->const_n = 1;
				break;
			}
			case 'k':{
				th->dim_k = (int64_t)atoi(optarg);
				th->const_k = 1;
				break;
			}
			case 'r':{
				char *str, *c_start, *c_stop, *c_step;
				char par[] = ":";
				str = optarg;
				c_start = strtok (str, par);
				c_stop  = strtok (NULL, par);
				c_step	= strtok (NULL, par);

				if (strcmp (c_start, "max") == 0)
					th->dim_start	= -888;
				else 
					th->dim_start	= atoi (c_start);

				if (strcmp (c_stop, "max") == 0)
					th->dim_stop	= -888;
				else 
					th->dim_stop	= atoi (c_stop);

				if (strcmp (c_step, "log") == 0)
					th->dim_step	= -222;
				else 
					th->dim_step	= atoi (c_step);

				break;
			}
			// options
			case 'a':{
				// N, T, C
				th->tranA = optarg[0];
				break;
			}
			case 'b':{
				// N, T, C
				th->tranB = optarg[0];
				break;
			}
			case 'u':{
				// U, L
				th->uplo = optarg[0];
				break;
			}
			case 'd':{
				// N, U
				th->diag = optarg[0];
				break;
			}
			case 's':{
				// L, R
				th->side = optarg[0];
				break;
			}
			// testing
			case 'o':{
				th->mode = optarg[0];
				break;
			}
			case 'v':{
				th->kernelnumber = atoi(optarg);
				break;
			}
			case 't':{
				th->cublasatomic = 1;
				break;
			}
			case 'x':{
				th->Tx_B_static = atoi(optarg);
				break;
			}
			case 'y':{
				th->Ty_B_static = atoi(optarg);
				break;
			}
			case 'p':{
				th->test_points = atoi(optarg);
				break;
			}
			case 'h':{
				printf ("testing program:\n");
				break;
			}
			case 'i':{
				th->incx = atoi(optarg);
				break;
			}
			case 'j':{
				th->incy = atoi(optarg);
				break;
			}
		}
	}
	
// ----- problem setting -----
	int64_t max_size = 0;
	int64_t dim_1, dim_2;

	switch (th->const_n + th->const_m + th->const_k) {
		case 0: {
			max_size = ROUTINE_MAXSIZE (th->devprop.totalGlobalMem, MEMMARGIN, th->word_size);
			break;
		}
		case 1: {
			dim_1 = th->dim_m + th->dim_n + th->dim_k;
			max_size = ROUTINE_MAXSIZE_1 (th->devprop.totalGlobalMem, MEMMARGIN, th->word_size, dim_1);
			break;
		}
		case 2: {
			if (th->const_m == 0) {
				dim_1 = th->dim_n;
				dim_2 = th->dim_k;
			} else if (th->const_n == 0) {
				dim_1 = th->dim_m;
				dim_2 = th->dim_k;
			} else {
				dim_1 = th->dim_n;
				dim_2 = th->dim_m;
			}
			max_size = ROUTINE_MAXSIZE_2 (th->devprop.totalGlobalMem, MEMMARGIN, th->word_size, dim_1, dim_2);
			break;
		}
	}

	if (th->dim_start == -888) th->dim_start = max_size;
	if (th->dim_stop  == -888) th->dim_stop = max_size;
	if (th->test_points != 0) {
		int32_t ccc = 128, ttt = 1;
		ttt = (th->dim_stop/(th->test_points)/ccc)*ccc;
		while (ttt < 1 && ccc > 0) {
			ccc /= 2;
			ttt = (th->dim_stop/(th->test_points)/ccc)*ccc;
		}
		if (ttt < 1) ttt = 1;
		th->dim_step = ttt;
		th->dim_start = th->dim_step;
	}
	
	th->test_points = (th->dim_stop-th->dim_start)/th->dim_step;

// ---------------------
		
}

void set_mublas_handler (testingHandle_t* th, mublasHandle_t* mh)
{
	mh->Tx_B_static = th->Tx_B_static;
	mh->Ty_B_static = th->Ty_B_static;
	mh->kernelnumber = th->kernelnumber;
}

// ------------------------------------------------------------------------------------------------------

double
gettime () {
	struct timeval tv;
	cudaDeviceSynchronize ();
	gettimeofday (&tv, NULL);
	return tv.tv_sec + (double) tv.tv_usec * 1e-6;
}

// ------------------------------------------------------------------------------------------------------
// DF-operation

#define SPLITTER 536870913.0 // 2^29+1
cudfreal double_to_cudfreal (const double d)
{
	double t, h, l;
	t = SPLITTER * d;
	h = t - (t - d);
	l = d - h;
	return make_cudfreal ((float)h, (float)l);
}

double cudfreal_to_double (const cudfreal df)
{
	return (double)df.x + (double)df.y;
}

// ------------------------------------------------------------------------------------------------------
	
#if defined (MBLAS_CHECK) && (defined (PREC_W) || defined (PREC_U))
#define RAND qdrand()/((REFERENCE_BASETYPE)RAND_MAX+1.)
#else
#define RAND rand()/((REFERENCE_BASETYPE)RAND_MAX+1.)
#endif

void
init_vector (uint64_t n, GPU_TYPE *x) {
	int32_t sig = 1;
	for (uint64_t i = 0; i < n; i++) {
		if (rand()%2 == 0) sig *= -1;
		REFERENCE_BASETYPE r1 = sig*RAND;
#if defined (PREC_C) || defined (PREC_Z) || defined (PREC_U) || defined (PREC_V)
		if (rand()%2 == 0) sig *= -1;
		REFERENCE_BASETYPE r2 = sig*RAND;
		#if defined (PREC_U) || !defined (MBLAS_CHECK)
		x[i] = MAKE_GPU_TYPE (r1, r2);
		#else
		x[i] = MAKE_GPU_TYPE (r1.x[0], r2.x[0]);
		#endif
#else
		#if defined (PREC_W) || !defined (MBLAS_CHECK)
		x[i] = MAKE_GPU_TYPE (r1);
		#else
		x[i] = MAKE_GPU_TYPE (r1.x[0]);
		#endif
#endif
	}
}

#ifdef MBLAS_CHECK
void
copy_vector (uint64_t n, GPU_TYPE *src, MBLAS_TYPE *dst) {
	for (uint64_t i = 0; i < n; i++) {
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_DB)
		dst[i] = qd_real ((double)src[i], 0., 0., 0.);
#endif
#if defined (PREC_W)
		dst[i] = qd_real (src[i].x, src[i].y, 0., 0.);
#endif
#if defined (PREC_C) || defined (PREC_Z) 
		dst[i] = qd_complex (qd_real ((double)src[i].x, 0., 0., 0.), qd_real ((double)src[i].y, 0., 0., 0.));
#endif
#if defined (PREC_U) 
		dst[i] = qd_complex (qd_real (src[i].x.x, src[i].x.y, 0., 0.), qd_real (src[i].y.x, src[i].y.y, 0., 0.));
#endif
#if defined (PREC_B)
		dst[i] = qd_real (cudfreal_to_double (src[i]), 0., 0., 0.);
#endif
#if defined (PREC_V) 
		dst[i] = qd_complex (qd_real (cudfreal_to_double (src[i].x), 0., 0., 0.), qd_real (cudfreal_to_double (src[i].y), 0., 0., 0.));
#endif
	}
}
#else
void
copy_vector (uint64_t n, GPU_TYPE *src, CBLAS_TYPE *dst) {
	for (uint64_t i = 0; i < n; i++) {
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_DB)
		dst[i] = src[i];
#endif
#if defined (PREC_W)
		dst[i] = src[i].x;
#endif
#if defined (PREC_C) || defined (PREC_Z) 
		dst[i].r = src[i].x;
		dst[i].i = src[i].y;
#endif
#if defined (PREC_U) 
		dst[i].r = src[i].x.x;
		dst[i].i = src[i].y.x;
#endif
#if defined (PREC_B)
		dst[i] = cudfreal_to_double (src[i]);
#endif
#if defined (PREC_V)
		dst[i].r = cudfreal_to_double (src[i].x);
		dst[i].i = cudfreal_to_double (src[i].y);
#endif
	}
}
#endif

// check routine (|trg-ref|/|ref|) -------------------------------------------------
void
check_result (uint64_t m, GPU_TYPE *target, REFERENCE_TYPE *ref) {
	int64_t size = sizeof (REFERENCE_TYPE);
	REFERENCE_TYPE negone = REFERENCE_NEGONE;
	REFERENCE_TYPE *target_w, *work;
	REFERENCE_BASETYPE Tnrm, Rnrm, rerr;

	work = (REFERENCE_TYPE *) malloc (size * m);
	target_w = (REFERENCE_TYPE *) malloc (size * m);
	copy_vector (m, target, target_w);
	Rnrm = REFERENCE_NRM (m, ref, 1);
#if !defined (MBLAS_CHECK) && ( defined (PREC_C) || defined (PREC_Z) || defined (PREC_U) || defined (PREC_V) )
	REFERENCE_AXPY (m, &negone, ref, 1, target_w, 1);
#else
	REFERENCE_AXPY (m, negone, ref, 1, target_w, 1);
#endif
	Tnrm = REFERENCE_NRM (m, target_w, 1);
	rerr = Tnrm / Rnrm;
#ifdef MBLAS_CHECK
	printf("\t%1.3e\t(lo:\t%1.3e)\t", rerr.x[0], rerr.x[1]);
#else
	printf("\t%1.3e\t", rerr);
#endif
	free (target_w);
	free (work);
}
// check routine (|trg-ref|/|ref|) -------------------------------------------------

// print routine -------------------------------------------------
void
print_mublas_kernel_info (mublasHandle_t* mh)
{
#ifndef NODISPINFO
	printf ("\t%d\t%d\t%d\t%d\t%d\t%1.3f\t%1.3f\t%1.3f\t%d", mh->Tx_B, mh->Ty_B, mh->S_B, mh->B_D, mh->B_M_max, mh->BO, mh->WO, mh->BO*mh->WO, mh->kernelnumber);
#endif
}

void
print_info (testingHandle_t* th)
{
#ifndef NODISPINFO
	printf ("%% ------------------------------------------\n");
	system("echo -n '% Date\t'; date");
	system("echo -n '% Host\t'; hostname");
	printf ("%% Device info ------------------------------\n");
	printf("%% \tDevice\t%s\n", th->devprop.name);
	printf("%% \tTheoretical Peak Memory Bandwidth\t%3.1f [GB/s]\n", th->device_bytes_per_sec * 1.e-9);
	printf("%% \tTheoretical Peak Performance (SP)\t%4.1f [GFlops]\n", th->device_flops_per_sec_sp * 1.e-9);
	printf("%% \tTheoretical Peak Performance (DP)\t%4.1f [GFlops]\n", th->device_flops_per_sec_dp * 1.e-9);
	printf("%% \tTotal Gmem\t%4.1f [GB]\n", th->devprop.totalGlobalMem * 1e-9);
	printf("%% \tCC\t%d.%d\n", th->devprop.major, th->devprop.minor);
	printf("%% \tECC Enabled\t%d\n", th->devprop.ECCEnabled);
	printf("%% \tMultiProcessorCount\t%d\n", th->devprop.multiProcessorCount);
	printf("%% \tDriver\t%d\n", th->driverVersion);
	printf("%% \tRuntime\t%d\n", th->runtimeVersion);
	printf ("%% ------------------------------------------\n");
#endif
}

void
print_info2 (testingHandle_t* th)
{
#ifndef NODISPINFO
	printf ("%% Testing info -----------------------------\n");
	printf ("%%\tRoutine\t");
#ifdef GEMV
	#ifdef PREC_S
	printf ("SGEMV\n");
	#elif PREC_D
	printf ("DGEMV\n");
	#elif PREC_C
	printf ("CGEMV\n");
	#elif PREC_Z
	printf ("ZGEMV\n");
	#elif PREC_W
	printf ("WGEMV\n");
	#elif PREC_U
	printf ("UGEMV\n");
	#elif PREC_B
	printf ("BGEMV\n");
	#elif PREC_V
	printf ("VGEMV\n");
	#endif
#endif
#ifdef SYMV
	#ifdef PREC_S
	printf ("SSYMV\n");
	#elif PREC_D
	printf ("DSYMV\n");
	#elif PREC_W
	printf ("WSYMV\n");
	#elif PREC_B
	printf ("BSYMV\n");
	#endif
#endif
#ifdef HEMV
	#ifdef PREC_C
	printf ("CHEMV\n");
	#elif PREC_Z
	printf ("ZHEMV\n");
	#elif PREC_U
	printf ("UHEMV\n");
	#elif PREC_V
	printf ("VHEMV\n");
	#endif
#endif


	printf("%% \tGmem allocation\t");
	#ifdef PITCH
	printf("pitched\n");
	#else
	printf("non-pitched\n");
	#endif
	printf ("%%\tMINLOOP\t%d\n", MINLOOP);
	printf ("%%\tMAXLOOP\t%d\n", MAXLOOP);
	printf ("%%\tMINSEC\t%f\n", MINSEC);

	printf ("%%\tLibrary\t");
	#ifdef MUBLAS 
	printf ("MUBLAS\n");
	#elif CUBLAS
	printf ("CUBLAS\n");
	#elif KBLAS
	printf ("KBLAS\n");
	#elif MAGMA
	printf ("MAGMA\n");
	#elif CULA
	printf ("CULA\n");
	#elif ASPEN
	printf ("ASPEN.K2\n");
	#else
	printf ("ERROR\n");
	#endif

	if (th->mode == 'c') {
		#ifdef MBLAS_CHECK
		printf ("%%\tCheck\tMBLAS\n");
		#else
		printf ("%%\tCheck\tCBLAS\n");
		#endif
	}
	printf ("%% ------------------------------------------\n");
	printf ("%% BLAS parameter ---------------------------\n");
	printf ("%%\ttranA\t%c\n", th->tranA);
	printf ("%%\ttranB\t%c\n", th->tranB);
	printf ("%%\tuplo\t%c\n", th->uplo);
	printf ("%%\tdiag\t%c\n", th->diag);
	printf ("%%\tside\t%c\n", th->side);
	printf ("%%\tdim_start\t%" PRId64 "\n", th->dim_start);
	printf ("%%\tdim_stop\t%" PRId64 "\n", th->dim_stop);
	printf ("%%\tdim_step\t%" PRId64 "\n", th->dim_step);
	printf ("%%\tM\t%" PRId64 "\n", th->dim_m);
	printf ("%%\tN\t%" PRId64 "\n", th->dim_n);
	printf ("%%\tK\t%" PRId64 "\n", th->dim_k);
	printf ("%%\tincx\t%" PRId64 "\n", th->incx);
	printf ("%%\tincy\t%" PRId64 "\n", th->incy);
	printf ("%%\tcublasatomic\t%d\n", th->cublasatomic);
	printf ("%% ------------------------------------------\n");
#endif
}

void
print_info3 (mublasHandle_t* mh)
{
#ifndef NODISPINFO
	printf ("%% MUBLAS info ------------------------------\n");
	printf ("%% \tTx_B_static\t%d\n", mh->Tx_B_static);
	printf ("%% \tTy_B_static\t%d\n", mh->Ty_B_static);
	printf ("%% \tbinaryVersion\t%d\n", mh->funcatb.binaryVersion);
	printf ("%% \tptxVersion\t%d\n", mh->funcatb.ptxVersion);
	printf ("%% ------------------------------------------\n");
#endif
}

void
print_devprop (cudaDeviceProp* devprop)
{
#ifndef NODISPINFO
	printf ("totalGlobalMem\t%zd\n", devprop->totalGlobalMem);
	printf ("sharedMemPerBlock\t%zd\n", devprop->sharedMemPerBlock);
	printf ("regsPerBlock\t%d\n", devprop->regsPerBlock);
	printf ("warpSize\t%d\n", devprop->warpSize);
	printf ("memPitch\t%zd\n", devprop->memPitch);
	printf ("maxThreadsPerBlock\t%d\n", devprop->maxThreadsPerBlock);
	printf ("maxThreadsDim[0]\t%d\n", devprop->maxThreadsDim[0]);
	printf ("maxThreadsDim[1]\t%d\n", devprop->maxThreadsDim[1]);
	printf ("maxThreadsDim[2]\t%d\n", devprop->maxThreadsDim[2]);
	printf ("maxGridSize[0]\t%d\n", devprop->maxGridSize[0]);
	printf ("maxGridSize[1]\t%d\n", devprop->maxGridSize[1]);
	printf ("maxGridSize[2]\t%d\n", devprop->maxGridSize[2]);
	printf ("clockRate\t%d\n", devprop->clockRate);
	printf ("totalConstMem\t%zd\n", devprop->totalConstMem);
	printf ("major\t%d\n", devprop->major);
	printf ("minor\t%d\n", devprop->minor);
	printf ("textureAlignment\t%zd\n", devprop->textureAlignment);
	printf ("texturePitrchAlignment\t%zd\n", devprop->texturePitchAlignment);
	printf ("deviceOverlap\t%d\n", devprop->deviceOverlap);
	printf ("multiProcessorCount\t%d\n",  devprop->multiProcessorCount);
	printf ("kernelExecTimeoutEnabled\t%d\n", devprop->kernelExecTimeoutEnabled);
	printf ("integrated\t%d\n", devprop->integrated);
	printf ("canMapHostMemory\t%d\n", devprop->canMapHostMemory);
	printf ("computeMode\t%d\n", devprop->computeMode);
	printf ("maxTexture1D\t%d\n", devprop->maxTexture1D);
	printf ("maxTexture1DMipmap\t%d\n", devprop->maxTexture1DMipmap);
	printf ("maxTexture1DLinear\t%d\n", devprop->maxTexture1DLinear);
	printf ("maxTexture2D[0]\t%d\n", devprop->maxTexture2D[0]);
	printf ("maxTexture2D[1]\t%d\n", devprop->maxTexture2D[1]);
	printf ("maxTexture2DMipmap[0]\t%d\n", devprop->maxTexture2DMipmap[0]);
	printf ("maxTexture2DMipmap[1]\t%d\n", devprop->maxTexture2DMipmap[1]);
	printf ("maxTexture2DLinear[0]\t%d\n", devprop->maxTexture2DLinear[0]);
	printf ("maxTexture2DLinear[1]\t%d\n", devprop->maxTexture2DLinear[1]);
	printf ("maxTexture2DLinear[2]\t%d\n", devprop->maxTexture2DLinear[2]);
	printf ("maxTexture2DGather[0]\t%d\n", devprop->maxTexture2DGather[0]);
	printf ("maxTexture2DGather[1]\t%d\n", devprop->maxTexture2DGather[1]);
	printf ("maxTexture3D[0]\t%d\n", devprop->maxTexture3D[0]);
	printf ("maxTexture3D[1]\t%d\n", devprop->maxTexture3D[1]);
	printf ("maxTexture3D[2]\t%d\n", devprop->maxTexture3D[2]);
	printf ("maxTexture3DAlt[0]\t%d\n", devprop->maxTexture3DAlt[0]);
	printf ("maxTexture3DAlt[1]\t%d\n", devprop->maxTexture3DAlt[1]);
	printf ("maxTexture3DAlt[2]\t%d\n", devprop->maxTexture3DAlt[2]);
	printf ("maxTextureCubemap\t%d\n", devprop->maxTextureCubemap);
	printf ("maxTexture1DLayered[0]\t%d\n", devprop->maxTexture1DLayered[0]);
	printf ("maxTexture1DLayered[1]\t%d\n", devprop->maxTexture1DLayered[1]);
	printf ("maxTexture2DLayered[0]\t%d\n", devprop->maxTexture2DLayered[0]);
	printf ("maxTexture2DLayered[1]\t%d\n", devprop->maxTexture2DLayered[1]);
	printf ("maxTexture2DLayered[2]\t%d\n", devprop->maxTexture2DLayered[2]);
	printf ("maxTextureCubemapLayered[0]\t%d\n", devprop->maxTextureCubemapLayered[0]);
	printf ("maxTextureCubemapLayered[1]\t%d\n", devprop->maxTextureCubemapLayered[1]);
	printf ("maxSurface1D\t%d\n", devprop->maxSurface1D);
	printf ("maxSurface2D[0]\t%d\n", devprop->maxSurface2D[0]);
	printf ("maxSurface2D[1]\t%d\n", devprop->maxSurface2D[1]);
	printf ("maxSurface3D[0]\t%d\n", devprop->maxSurface3D[0]);
	printf ("maxSurface3D[1]\t%d\n", devprop->maxSurface3D[1]);
	printf ("maxSurface3D[2]\t%d\n", devprop->maxSurface3D[2]);
	printf ("maxSurface1DLayered[0]\t%d\n", devprop->maxSurface1DLayered[0]);
	printf ("maxSurface1DLayered[1]\t%d\n", devprop->maxSurface1DLayered[1]);
	printf ("maxSurface2DLayered[0]\t%d\n", devprop->maxSurface2DLayered[0]);
	printf ("maxSurface2DLayered[1]\t%d\n", devprop->maxSurface2DLayered[1]);
	printf ("maxSurface2DLayered[2]\t%d\n", devprop->maxSurface2DLayered[2]);
	printf ("maxSurfaceCubemap\t%d\n", devprop->maxSurfaceCubemap);
	printf ("maxSurfaceCubemapLayered[0]\t%d\n", devprop->maxSurfaceCubemapLayered[0]);
	printf ("maxSurfaceCubemapLayered[1]\t%d\n", devprop->maxSurfaceCubemapLayered[1]);
	printf ("surfaceAlignment\t%zd\n", devprop->surfaceAlignment);
	printf ("concurrentKernels\t%d\n", devprop->concurrentKernels);
	printf ("ECCEnabled\t%d\n", devprop->ECCEnabled);
	printf ("pciBusID\t%d\n", devprop->pciBusID);
	printf ("pciDeviceID\t%d\n", devprop->pciDeviceID);
	printf ("pciDomainID\t%d\n", devprop->pciDomainID);
	printf ("tccDriver\t%d\n", devprop->tccDriver);
	printf ("asyncEngineCount\t%d\n", devprop->asyncEngineCount);
	printf ("unifiedAddressing\t%d\n", devprop->unifiedAddressing);
	printf ("memoryClockRate\t%d\n", devprop->memoryClockRate);
	printf ("memoryBusWidth\t%d\n", devprop->memoryBusWidth);
	printf ("l2CacheSize\t%d\n", devprop->l2CacheSize);
	printf ("maxThreadsPerMultiProcessor\t%d\n", devprop->maxThreadsPerMultiProcessor);
	printf ("streamPrioritiesSupported\t%d\n", devprop->streamPrioritiesSupported);
	printf ("globalL1CacheSupported\t%d\n", devprop->globalL1CacheSupported);
	printf ("localL1CacheSupported\t%d\n", devprop->localL1CacheSupported);
	printf ("sharedMemPerMultiprocessor\t%zd\n", devprop->sharedMemPerMultiprocessor);
	printf ("regsPerMultiprocessor\t%d\n", devprop->regsPerMultiprocessor);
//	printf ("managedMemSupported\t%d\n", devprop->managedMemSupported);
	printf ("isMultiGpuBoard\t%d\n", devprop->isMultiGpuBoard);
	printf ("multiGpuBoardGroupID\t%d\n", devprop->multiGpuBoardGroupID);
	printf ("----------------\n");
#endif
}
// print routine -------------------------------------------------

